package kmsHandler

import (
	"fmt"

	"code.fbi.h-da.de/danet/gnmi-target/handler"
	"code.fbi.h-da.de/danet/quant/goKMS/kms"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/event"
	gnmitargetygot "code.fbi.h-da.de/danet/quant/goKMS/model"
	"github.com/openconfig/gnmi/proto/gnmi"
	"github.com/openconfig/ygot/ygot"
	log "github.com/sirupsen/logrus"
)

type KeyStoreHandler struct {
	handler.DefaultPathHandler
	kms                     *kms.KMS
	events                  <-chan event.Event
	maxKeyFillLevelsDefined map[string]uint64
}

func NewKeyStoreHandler(kms *kms.KMS, maxKeyFillLevelsDefined map[string]uint64) *KeyStoreHandler {
	return &KeyStoreHandler{
		DefaultPathHandler: handler.DefaultPathHandler{
			Name: "key-store-handler",
			Paths: map[string]struct{}{
				"key-stores": {},
			},
		},
		kms:                     kms,
		maxKeyFillLevelsDefined: maxKeyFillLevelsDefined,
	}
}

func (yh *KeyStoreHandler) Init(config *handler.Config, publishToSubsFunc func([]*gnmi.Notification) error) error {
	yh.Config = config
	yh.PublishToSubs = publishToSubsFunc

	var err error
	yh.events, err = event.NewEventBus().Subscribe(event.KEY_STORE)
	if err != nil {
		return err
	}

	if _, err := yh.updateOrCreateKeyStores(); err != nil {
		return err
	}

	// Start the go routine that takes care of any update from the kms
	go func() {
		for {
			select {
			case <-yh.events:
				log.Println("Update for key stores.")

				diff, err := yh.updateOrCreateKeyStores()
				if err != nil {
					log.Errorf("Error within key stores subscription goroutine; %v", err)
					break
				}
				if err := yh.PublishToSubs(diff); err != nil {
					log.Errorf("Error within key stores subscription goroutine; %v", err)
				}
			}
		}
	}()

	return nil
}

func (yh *KeyStoreHandler) Update(c ygot.ValidatedGoStruct, jobs []*gnmi.Update) error {
	return nil
}

func (yh *KeyStoreHandler) updateOrCreateKeyStores() ([]*gnmi.Notification, error) {
	yh.Config.Lock()
	defer yh.Config.Unlock()

	copyCurrentConfig, err := ygot.DeepCopy(yh.Config.Data)
	if err != nil {
		return nil, err
	}

	newConfig, ok := copyCurrentConfig.(*gnmitargetygot.Gnmitarget)
	if !ok {
		return nil, fmt.Errorf("Wrong type, exptected: %T, got: %T", (*gnmitargetygot.Temp_KeyStores)(nil), copyCurrentConfig)
	}

	kmsPeers := yh.kms.PeersDeepCopy()

	confKeyStores := newConfig.GetOrCreateKeyStores()

	for peerKmsId, peer := range kmsPeers {
		confKeyStoreContainer := confKeyStores.GetOrCreateKeyStore(peerKmsId)

		confKeyStoreContainer.KmsPeerId = ygot.String(peerKmsId)
		confKeyStore := confKeyStoreContainer.GetOrCreateKeyStore()

		confKeyStore.KeyFillLevel = ygot.Uint64(uint64(peer.QuantumModule().KeyStore().Length()))
		maxKeyFillLevel, ok := yh.maxKeyFillLevelsDefined[peerKmsId]
		if !ok {
			return nil, fmt.Errorf("no max key fill level available for store with ID: %s", peerKmsId) // TODO: check if really want to return here
		}

		confKeyStore.MaxKeyFillLevel = ygot.Uint64(maxKeyFillLevel)
	}

	// validate struct
	if err := newConfig.Validate(); err != nil {
		return nil, err
	}

	notifications, err := ygot.DiffWithAtomic(yh.Config.Data, newConfig)
	if err != nil {
		return nil, err
	}

	yh.Config.Data = newConfig

	return notifications, nil
}
