package kms

import (
	"context"
	"encoding/base64"
	"fmt"
	"sync"
	"time"

	"github.com/google/uuid"
	log "github.com/sirupsen/logrus"

	etsi14 "code.fbi.h-da.de/danet/quant/etsi014/go/rest/etsi/client"
	pb "code.fbi.h-da.de/danet/quant/goKMS/api/gen/proto/go/kmsintercom"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/crypto"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/event"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/peers"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/store"
	kmstls "code.fbi.h-da.de/danet/quant/goKMS/kms/tls"
	"google.golang.org/grpc"
	"google.golang.org/grpc/codes"
	"google.golang.org/grpc/peer"
	"google.golang.org/grpc/status"
)

const etsi014RequestID = "etsi014"

type kmsTalkerServer struct {
	pb.UnimplementedKmsTalkerServer
	keyNegotationMutex sync.Mutex
	keyNegotiationMap  map[uuid.UUID]*store.KmsKSElement
	kms                *KMS
}

func (s *kmsTalkerServer) InterComCapabilities(ctx context.Context, in *pb.InterComCapabilitiesRequest) (capReply *pb.InterComCapabilitiesReply, err error) {
	log.Debugf("Received InterComCapabilities request from peer KMS: %v", in.GetKmsId())

	peerKmsId := in.GetKmsId()
	peer, err := s.kms.FindPeerById(peerKmsId)
	if err != nil {
		return nil, status.Errorf(codes.Aborted, err.Error(), in.GetKmsId())
	}

	if in.ResetKeyStore {
		log.Debugf("Resetting quantum module: %s for peer: %s", peer.QuantumModule().ID(), peerKmsId)
		peer.QuantumModule().Reset()
		log.Debugf("Initializing quantum module: %s for peer: %s", peer.QuantumModule().ID(), peerKmsId)
		if err := peer.QuantumModule().Initialize(); err != nil {
			return nil, status.Errorf(codes.Aborted, err.Error(), in.GetKmsId())
		}
		log.Debugf("Successfully initialized quantum module: %s for peer %s", peer.QuantumModule().ID(), peerKmsId)
	}

	// NOTE: InterComCapabilities should return the capabilities of the kms.
	// This could include supported key sizes, crypto algorithms, etc.
	// Therefore the proto definitions should be extended accordingly.

	return &pb.InterComCapabilitiesReply{
		Timestamp:   time.Now().Unix(),
		PeerKmsName: s.kms.kmsName,
	}, nil
}

func (s *kmsTalkerServer) KeyIdNotification(ctx context.Context, in *pb.KeyIdNotificationRequest) (*pb.KeyIdNotificationResponse, error) {
	peer, err := s.kms.FindPeerById(in.GetKmsId())
	if err != nil {
		return nil, status.Errorf(codes.Internal, "peer with ID: %s does not exist in peers", in.GetKmsId())
	}

	eqm, ok := peer.QuantumModule().(*peers.ETSI014HTTPQuantumModule)
	if !ok {
		return nil, status.Error(codes.Internal, "expected etsi014 quantum module")
	}

	switch {
	case !eqm.IsActive():
		log.Debugf("The key store for quantum module: %s is not active and denied incoming key sync attempts, current key store length: %d", eqm.ID(), eqm.KeyStore().Length())
		return nil, status.Errorf(codes.Aborted, "The corresponding key store is not active and does not accept incoming key sync attempts")
	case eqm.KeyStore().Length() >= int(eqm.MaxKeyFillLevel()):
		log.Debugf("The key store for quantum module: %s is at its maximum key fill level and does not accept incoming key sync attempts", eqm.ID())
		return nil, status.Errorf(codes.Aborted, "The corresponding key store is at its maximum key fill level and does not accept incoming key sync attempts")
	}

	etsi14KeyIds := make([]etsi14.KeyIDsRequestKeyIDsInner, len(in.KeyIds))
	for i, keyid := range in.KeyIds {
		etsi14KeyIds[i] = etsi14.KeyIDsRequestKeyIDsInner{
			KeyID: keyid,
		}
	}

	keyContainer, err := eqm.GetKeyWithIds(etsi14KeyIds)
	if err != nil {
		return nil, err
	}

	if err := store.AddETSIKeysToKeystore(eqm.KeyStore(), keyContainer.GetKeys()); err != nil {
		return nil, status.Errorf(codes.Internal, "failed to add keys to keystore; %s", err)
	}

	return &pb.KeyIdNotificationResponse{
		Timestamp: time.Now().Unix(),
	}, nil
}

func (s *kmsTalkerServer) SyncQkdBulk(ctx context.Context, in *pb.SyncQkdBulkRequest) (*pb.SyncQkdBulkResponse, error) {
	// NOTE: with "google.golang.org/grpc/peer" it would be possible to get the client ip directly
	p, _ := peer.FromContext(ctx)
	log.Infof("Received SyncQkdBulkRequest from %s", p.Addr.String())
	// check if a peer exists
	peer, err := s.kms.FindPeerById(in.GetKmsId())
	if err != nil {
		return nil, status.Errorf(codes.Internal, "peer with ID: %s does not exist in peers", in.GetKmsId())
	}

	eqm, ok := peer.QuantumModule().(*peers.DanetQuantumModule)
	if !ok {
		return nil, status.Errorf(codes.Internal, "expected emulated quantum module")
	}

	eqm.RawBulkKeysMutex.Lock()
	defer eqm.RawBulkKeysMutex.Unlock()
	for _, bulkId := range in.BulkId {
		if _, ok := eqm.RawBulkKeys[bulkId]; ok {
			return &pb.SyncQkdBulkResponse{
				Timestamp: time.Now().Unix(),
				BulkId:    bulkId,
			}, nil
		}
	}

	return nil, status.Errorf(codes.Internal, "Could not find a RawBulkKey that matches the provided IDs: %v", in.BulkId)
}

func (s *kmsTalkerServer) SyncKeyIdsForBulk(ctx context.Context, in *pb.SyncKeyIdsForBulkRequest) (*pb.SyncKeyIdsForBulkResponse, error) {
	peer, err := s.kms.FindPeerById(in.GetKmsId())
	if err != nil {
		return nil, status.Errorf(codes.Internal, "For KMS id: %s, no peer exists", in.GetKmsId())
	}

	eqm, ok := peer.QuantumModule().(*peers.DanetQuantumModule)
	if !ok {
		return nil, status.Errorf(codes.Internal, "expected a emulated quantum module")
	}

	eqm.RawBulkKeysMutex.Lock()
	defer eqm.RawBulkKeysMutex.Unlock()
	bulk, ok := eqm.RawBulkKeys[in.GetBulkId()]
	if !ok {
		return nil, status.Errorf(codes.Internal, "bulk key with id: %d not found", in.GetBulkId())
	}

	eqm.SetKeyStore(store.NewKmsKeyStore(256))
	_, keyData, err := eqm.KeyChopper(bulk, in.GetKeyId())
	if err != nil {
		return nil, status.Errorf(codes.Internal, "%v", err)
	}

	if err := eqm.KeyStore().AddKeys(keyData); err != nil {
		log.Error(err)
	}

	delete(eqm.RawBulkKeys, in.GetBulkId())

	// update the peer status to up
	peer.SetStatus(peers.KmsPeerUp)
	// Send notification about change
	if peer.EventBus() != nil {
		go func() {
			err := peer.EventBus().Publish(event.NewPeerEvent(peer.TcpSocketStr))
			if err != nil {
				log.Error(err)
			}
		}()
	}

	return &pb.SyncKeyIdsForBulkResponse{
		Timestamp: time.Now().Unix(),
	}, nil
}

func (s *kmsTalkerServer) InterComTransportKeyNegotiation(ctx context.Context, in *pb.InterComTransportKeyNegotiationRequest) (*pb.InterComTransportKeyNegotiationResponse, error) {
	// NOTE: For the current prototype it is assumed that a negotiation request
	// is always valid. In the future an incoming negotiation request should
	// also check for a suitable forwarding assignment from the controller.
	keyID, err := uuid.Parse(in.GetKeyToUse())
	if err != nil {
		return nil, status.Errorf(codes.InvalidArgument, "key with id: %s can not be parsed to uuid", in.GetKeyToUse())
	}
	pathId, err := uuid.Parse(in.GetPathID())
	if err != nil {
		return nil, status.Errorf(codes.InvalidArgument, "path id: %s can not be parsed to uuid", in.GetPathID())
	}

	route, ok := s.kms.routingTable[pathId]
	if !ok {
		return nil, status.Errorf(codes.Internal, "There is no route for the given pathID: %s .", in.PathID)
	}

	quantumModuleKeyStore := route.Previous.QuantumModule().KeyStore()

	key, err := quantumModuleKeyStore.GetKeyWithID(keyID)
	if err != nil {
		return nil, status.Errorf(codes.Internal, "%v", err)
	}

	s.keyNegotationMutex.Lock()
	defer s.keyNegotationMutex.Unlock()
	s.keyNegotiationMap[pathId] = key

	quantumModuleKeyStore.DeleteKey(keyID)

	return &pb.InterComTransportKeyNegotiationResponse{Timestamp: time.Now().Unix()}, nil
}

func (s *kmsTalkerServer) KeyForwarding(ctx context.Context, in *pb.KeyForwardingRequest) (*pb.KeyForwardingResponse, error) {
	pathId, err := uuid.Parse(in.GetPathId())
	if err != nil {
		return nil, status.Errorf(codes.InvalidArgument, "")
	}

	processId, err := uuid.Parse(in.GetProcessId())
	if err != nil {
		return nil, status.Errorf(codes.InvalidArgument, "")
	}

	keyID, err := uuid.Parse(in.GetKey().GetId())
	if err != nil {
		return nil, status.Errorf(codes.Internal, "%s", err)
	}

	route, ok := s.kms.routingTable[pathId]
	if !ok {
		return nil, status.Errorf(codes.Internal, "There is no route for the given pathID: %s .", in.PathId)
	}

	log.Debugf("%s received a key: %s, from %s", s.kms.kmsName, in.GetKey(), route.Previous.TcpSocketStr)

	s.keyNegotationMutex.Lock()
	defer s.keyNegotationMutex.Unlock()
	decryptKey, ok := s.keyNegotiationMap[pathId]
	if !ok {
		return nil, status.Errorf(codes.InvalidArgument, "")
	}

	cryptoAlgorithm, err := crypto.GetCryptoAlgorithmByName(in.GetCryptoAlgorithm().String())
	if err != nil {
		return nil, status.Errorf(codes.Internal, "%s", err)
	}

	decryptedKey, err := s.getDecryptedKey(decryptKey.Key, cryptoAlgorithm, in.GetKey())
	if err != nil {
		return nil, status.Errorf(codes.Internal, "%s", err)
	}

	if route.Next != nil {
		log.Debugf("%s forwards payload to : %s", s.kms.kmsName, route.Next.TcpSocketStr)

		go route.Next.SendPayload(&crypto.Key{ //nolint:errcheck
			ID:  keyID,
			Key: decryptedKey,
		}, pathId, processId)
	} else {
		log.Debugf("%s received the final payload: %s", s.kms.kmsName, decryptedKey)

		s.storeReceivedPlatformKey(route.RemoteKMS.Id, in.GetProcessId(), keyID, decryptedKey)

		remoteKMSAdrress := fmt.Sprintf("%s:%d", route.RemoteKMS.Address, route.RemoteKMS.Port)

		err = s.sendAcknowledgeKeyForwarding(ctx, remoteKMSAdrress, in.PathId, in.ProcessId, in.GetKey().GetId())
		if err != nil {
			return nil, err
		}
	}

	return &pb.KeyForwardingResponse{Timestamp: time.Now().Unix()}, nil
}

func (s *kmsTalkerServer) AckKeyForwarding(ctx context.Context, in *pb.AckKeyForwardingRequest) (*pb.AckKeyForwardingResponse, error) {
	pathId, err := uuid.Parse(in.GetPathId())
	if err != nil {
		return nil, status.Errorf(codes.InvalidArgument, "")
	}

	// TODO: additional checks, like e.g.:
	// - Did this kms initiate the key forwarding?
	// - Are pathId and processId valid?
	// - Is the keyId valid?

	err = s.kms.receiver.InformReceiver(pathId)
	if err != nil {
		return nil, status.Errorf(codes.InvalidArgument, "Failed while informing Receiver; err: %v", err)
	}

	return &pb.AckKeyForwardingResponse{Timestamp: time.Now().Unix()}, nil
}

func (s *kmsTalkerServer) KeyDelivery(ctx context.Context, in *pb.KeyDeliveryRequest) (*pb.KeyDeliveryResponse, error) {
	// parse path id
	keyId, err := uuid.Parse(in.GetKeyId())
	if err != nil {
		return nil, status.Errorf(codes.Internal, "%s", err)
	}

	// look up PK
	platformKey, err := s.kms.GetSpecificPlatformKey(in.GetKmsId(), keyId)
	if err != nil {
		return nil, status.Errorf(codes.NotFound, "%s", err)
	}

	// decrypt keys
	akmsKSAKeys := make([]crypto.KSAKey, len(in.Keys))
	for i, key := range in.Keys {
		// decrypt the key
		cryptoAlgorithm, err := crypto.GetCryptoAlgorithmByName(in.GetCryptoAlgorithm().String())
		if err != nil {
			return nil, status.Errorf(codes.Internal, "%s", err)
		}
		decryptedKSAKey, err := s.getDecryptedKey(platformKey.Value, cryptoAlgorithm, key)
		if err != nil {
			return nil, status.Errorf(codes.Internal, "%s", err)
		}

		// base64 encode
		ksaKeyAsString := base64.StdEncoding.EncodeToString(decryptedKSAKey)

		keyId := key.GetId()

		log.Debugf("KeyID: %s, Key: %s", keyId, ksaKeyAsString)

		akmsKSAKeys[i] = crypto.KSAKey{
			KeyID: keyId,
			Key:   ksaKeyAsString,
		}
	}

	if s.kms.keyStoreChannel != nil && in.GetRequestId() == etsi014RequestID {
		log.Debugf("(ETSI14) Pushing to KeyStoreChannel: %v in %s", s.kms.keyStoreChannel, s.kms.kmsName)
		s.kms.keyStoreChannel <- akmsKSAKeys
	} else if s.kms.ckmsAkmsClient != nil {
		go s.kms.ckmsAkmsClient.SendKSAKeysToRequestingInstances(in.GetRequestId(), platformKey.ProcessId, akmsKSAKeys) //nolint:errcheck
	}

	return &pb.KeyDeliveryResponse{Timestamp: time.Now().Unix()}, nil
}

func (s *kmsTalkerServer) getDecryptedKey(keyForDecryption []byte, cryptoAlgorithm crypto.CryptoAlgorithm, encryptedKey *pb.Key) ([]byte, error) {
	keyAsByte, err := base64.StdEncoding.DecodeString(encryptedKey.GetKey())
	if err != nil {
		return nil, err
	}

	nonceAsByte, err := base64.StdEncoding.DecodeString(encryptedKey.GetNonce())
	if err != nil {
		return nil, err
	}

	decryptedKey, err := cryptoAlgorithm.Decrypt(nonceAsByte, keyAsByte, keyForDecryption)
	if err != nil {
		return nil, err
	}

	return decryptedKey, nil
}

func (s *kmsTalkerServer) storeReceivedPlatformKey(remoteKmsID, processID string, keyID uuid.UUID, decryptedKey []byte) {
	s.kms.PKStoreMutex.Lock()
	defer s.kms.PKStoreMutex.Unlock()

	keys, ok := s.kms.PKStore[remoteKmsID]
	if !ok {
		s.kms.PKStore[remoteKmsID] = map[uuid.UUID]*PlatformKey{
			keyID: {
				Id:        keyID,
				Value:     decryptedKey,
				ProcessId: processID,
			},
		}
	} else {
		keys[keyID] = &PlatformKey{
			Id:        keyID,
			Value:     decryptedKey,
			ProcessId: processID,
		}
	}

	log.Debug("Current PKSTORE: ", s.kms.PKStore)
}

func (s *kmsTalkerServer) sendAcknowledgeKeyForwarding(ctx context.Context, remoteKmsAddr, pathID, processID, keyID string) error {
	gRPCTransportCreds, err := kmstls.GenerateGRPCClientTransportCredsBasedOnTLSFlag(s.kms.tlsConfig)
	if err != nil {
		return fmt.Errorf("unable to generate gRPC transport creds: %w", err)
	}

	newPeerConn, err := grpc.NewClient(remoteKmsAddr, grpc.WithTransportCredentials(gRPCTransportCreds))
	if err != nil {
		return err
	}

	// inform about successful key forwarding
	client := pb.NewKmsTalkerClient(newPeerConn)

	_, err = client.AckKeyForwarding(ctx, &pb.AckKeyForwardingRequest{
		Timestamp: time.Now().Unix(),
		PathId:    pathID,
		ProcessId: processID,
		KeyId:     keyID,
	})
	if err != nil {
		return err
	}

	log.Debugf("Successfully send ACK for pathID: %s to %s", pathID, remoteKmsAddr)

	return nil
}
