package main

import (
	"flag"
	"os"
	"time"

	"code.fbi.h-da.de/danet/quant/quantumlayer"
	randomnumbergenerator "code.fbi.h-da.de/danet/quant/quantumlayer/randomNumberGenerator"
	replaynumbergenerator "code.fbi.h-da.de/danet/quant/quantumlayer/replayNumberGenerator"
	pb "code.fbi.h-da.de/danet/quipsec/gen/go/quipsec"
	"github.com/sirupsen/logrus"
	"google.golang.org/grpc"
	"google.golang.org/grpc/credentials/insecure"
	"google.golang.org/grpc/resolver"
	"gopkg.in/yaml.v3"
)

type Config struct {
	KMSAddr                     string `yaml:"KMSAddr"`                     // Address of KMS
	OperationMode               string `yaml:"OperationMode"`               // "normal" or "replay"
	SleepTimerForGenerationInMs int    `yaml:"SleepTimerForGenerationInMs"` // sleep timer for generation in ms, only relevant if OperationMode is "normal" (0 equals default=5000)
	ReplayFilePath              string `yaml:"replayFilePath"`              // path to the replay file, only relevant if OperationMode is "replay"
	Seed                        int64  `yaml:"Seed"`                        // seed for random number generator, if empty set to current day
}

func main() {
	configPath := flag.String("config", "", "path to the config file")
	logLevel := flag.String("log", "info", "logrus lof level (debug, info, warn, error, fatal, panic), default: info")
	noGRPCPassthrough := flag.Bool("noGRPCPassthrough", false, "set the default resolve scheme for grpc to not use passthrough, default is false")

	flag.Parse()

	if *noGRPCPassthrough {
		logrus.Info("gRPC default resolver scheme is not set to passthrough. This might cause issues with the gRPC connection when no real DNS server is available as each gRPC requests requires a DNS request.")
	} else {
		logrus.Info("Setting gRPC default resolver scheme to passthrough. No DNS queries are being made when doing a gRPC request.")
		resolver.SetDefaultScheme("passthrough")
	}

	if *configPath == "" {
		logrus.Fatal("path to config file required")
	}

	// parse string, this is built-in feature of logrus
	ll, err := logrus.ParseLevel(*logLevel)
	if err != nil {
		ll = logrus.InfoLevel
		logrus.Warn("Invalid log level, using default: ", ll)
	}

	// set global log level
	logrus.SetLevel(ll)
	logrus.Info("Setting log level to ", ll)

	// unmarshal config
	config := &Config{}
	file, err := os.ReadFile(*configPath)
	if err != nil {
		logrus.Fatal(err)
	}
	if err := yaml.Unmarshal(file, config); err != nil {
		logrus.Fatal(err)
	}

	if config.Seed == 0 {
		// If no seed set use current date as seed.
		// This is not secure at all, but this emulator is just for testing purposes.
		// It simply guarantees, that we get some different values each day, to see some 'refreshing' new numbers each day in the debugger.
		year, month, day := time.Now().Date()
		config.Seed = int64(year + int(month) + day)
	}

	if config.SleepTimerForGenerationInMs == 0 {
		config.SleepTimerForGenerationInMs = 5000
	}

	var randomNumberSource quantumlayer.RandomNumberSource
	if config.OperationMode == "replay" {
		if config.ReplayFilePath == "" {
			logrus.Fatal("Replay mode requires a replay file path")
		}
		logrus.Info("Replay mode enabled, replay file path: ", config.ReplayFilePath)
		randomNumberSource = replaynumbergenerator.NewReplayNumberGenerator(config.ReplayFilePath, config.Seed)
	} else {
		logrus.Info("Normal mode enabled")
		randomNumberSource = randomnumbergenerator.NewRandomNumberGenerator(900, 1400, config.Seed, 5000)
	}
	err = randomNumberSource.InitializeRandomGenerator()
	if err != nil {
		logrus.Fatalf("Failed to initialize random generator: %v", err)
	}

	const connectionRetries = 240
	// Connect to KMS
	logrus.Info("Connecting to KMS ", config.KMSAddr)
	var newPeerConn *grpc.ClientConn
	for i := 0; i < connectionRetries; i++ {
		newPeerConn, err = grpc.NewClient(config.KMSAddr, grpc.WithTransportCredentials(insecure.NewCredentials()))
		if err == nil {
			logrus.Info("Connected to KMS")
			break
		}
		if i == connectionRetries-1 {
			logrus.Fatalf("Failed to connect to KMS for %d times ; aborting. Original error message: %s", connectionRetries, err.Error())
			return
		}
		logrus.Errorf("Failed to connect to KMS with error: %s, retrying in 1 second (attempt %d/%d)", err.Error(), i+1, connectionRetries)
		time.Sleep(1 * time.Second)
	}

	kmsClient := pb.NewKmsQkdmCommunicationServiceClient(newPeerConn)

	ql := quantumlayer.NewQuantumlayerEmulator(kmsClient, randomNumberSource)
	ql.Start()
}
