package replaynumbergenerator

import (
	"encoding/csv"
	"math/rand"
	"os"
	"strconv"
	"time"

	"code.fbi.h-da.de/danet/quant/quantumlayer/utils"

	log "github.com/sirupsen/logrus"
)

// ReplayNumberGenerator is a struct that generates "random numbers" based on a replay file.
type ReplayNumberGenerator struct {
	currentIteration int
	pathToReplayFile string
	seed             int64
	ReplayData       ReplayData
	randomGenerator  *rand.Rand
}

// NewReplayNumberGenerator creates a new ReplayNumberGenerator.
func NewReplayNumberGenerator(pathToReplayFile string, seed int64) *ReplayNumberGenerator {
	return &ReplayNumberGenerator{
		pathToReplayFile: pathToReplayFile,
		seed:             seed,
		currentIteration: 0,
	}
}

// InitializeRandomNumbers initializes the random numbers based on the replay file.
func (rng *ReplayNumberGenerator) InitializeRandomGenerator() error {
	replayData, err := readReplayFile(rng.pathToReplayFile)
	if err != nil {
		return err
	}
	rng.ReplayData = replayData
	rng.randomGenerator = rand.New(rand.NewSource(rng.seed))
	rng.currentIteration = 0

	return nil
}

// GenerateRandomNumbers generates "random numbers" based on the replay file.
func (rng *ReplayNumberGenerator) GenerateRandomBytes() ([]byte, int, error) {
	timeToSleep := time.Duration(rng.ReplayData.TimeSinceLastEntryInMs[rng.currentIteration]) * time.Millisecond
	time.Sleep(time.Duration(timeToSleep))

	numberOfBytes := rng.ReplayData.NumberOfBytes[rng.currentIteration]
	data, err := utils.GetRandomDataForGivenLength(rng.randomGenerator, numberOfBytes)
	if err != nil {
		log.Errorf("Error generating random bytes: %s", err)
		return nil, numberOfBytes, err
	}

	rng.currentIteration++
	return data, numberOfBytes, nil
}

// GetReplayData returns the replay data from the provided replay file.
func readReplayFile(filePath string) (ReplayData, error) {
	f, err := os.Open(filePath)
	if err != nil {
		log.Errorf("Unable to open file %s, %s", filePath, err.Error())
		return ReplayData{}, err
	}
	defer f.Close() // nolint: errcheck

	csvReader := csv.NewReader(f)
	csvReader.Comma = ';' // Use semicolon as delimiter

	// Skip the header
	_, err = csvReader.Read()
	if err != nil {
		log.Errorf("Unable to read header for %s; %s", filePath, err.Error())
		return ReplayData{}, err
	}

	records, err := csvReader.ReadAll()
	if err != nil {
		log.Errorf("Unable to parse file as CSV for %s; %s", filePath, err.Error())
		return ReplayData{}, err
	}

	var replayData ReplayData

	for _, record := range records {
		timestamp, err := strconv.Atoi(record[0])
		if err != nil {
			log.Errorf("Unable to parse timestamp for record %s; %s ", record[0], err.Error())
			return replayData, err
		}
		replayData.TimeSinceLastEntryInMs = append(replayData.TimeSinceLastEntryInMs, timestamp)

		numberOfBits, err := strconv.Atoi(record[1])
		if err != nil {
			log.Errorf("Unable to parse number of bits for record %s; %s ", record[1], err.Error())
			return replayData, err
		}
		//check if number of bits is dividable by 8
		if numberOfBits%8 != 0 {
			log.Errorf("Number of bits is not dividable by 8 and therefore not directly valid as byte for record %s ", record[1])
			return replayData, err
		}
		// convert to bytes
		numberOfBytes := numberOfBits / 8

		replayData.NumberOfBytes = append(replayData.NumberOfBytes, numberOfBytes)
	}

	return replayData, nil
}

type ReplayData struct {
	ReplayFilePath         string
	TimeSinceLastEntryInMs []int
	NumberOfBytes          []int
}
